@@ -35,6 +35,7 @@ from iqm.models.playlist.segment import Segment as SC_Schedule  # iqm.models nam
 from iqm.models.playlist.waveforms import to_canonical
 
 from exa.common.qcm_data.chip_topology import ChipTopology, sort_components
+from iqm.pulse.base_utils import _dicts_differ, merge_dicts
 from iqm.pulse.gate_implementation import (
     CompositeCache,
     GateImplementation,
@@ -42,12 +43,14 @@ from iqm.pulse.gate_implementation import (
     OILCalibrationData,
     OpCalibrationDataTree,
 )
-from iqm.pulse.gates import _default_operations, _exposed_implementations
+from iqm.pulse.gates import _exposed_implementations
+from iqm.pulse.gates.default_gates import _default_implementations, _quantum_ops_library
 from iqm.pulse.playlist.channel import ChannelProperties, ProbeChannelProperties
 from iqm.pulse.playlist.instructions import (
     AcquisitionMethod,
     ComplexIntegration,
     ConditionalInstruction,
+    FluxPulse,
     Instruction,
     IQPulse,
     MultiplexedIQPulse,
@@ -69,7 +72,12 @@ from iqm.pulse.scheduler import (
     extend_schedule_new,
 )
 from iqm.pulse.timebox import SchedulingAlgorithm, SchedulingStrategy, TimeBox
-from iqm.pulse.utils import load_yaml, merge_dicts
+from iqm.pulse.utils import (
+    _process_implementations,
+    _validate_locus_defaults,
+    _validate_op_attributes,
+    load_yaml,
+)
 
 logger = logging.getLogger(__name__)
 
@@ -139,59 +147,6 @@ def load_config(path: str) -> tuple[QuantumOpTable, OpCalibrationDataTree]:
     return build_quantum_ops(ops), calibration
 
 
-def build_quantum_ops(ops: dict[str, Any]) -> QuantumOpTable:
-    """Builds the table of known quantum operations.
-
-    Hardcoded default native ops table is extended by the ones in `ops`.
-    In case of name collisions, the content of `ops` takes priority over the defaults.
-
-    Args:
-        ops: Contents of the ``gate_definitions`` section defining the quantum operations in the
-            configuration YAML file. Modified by the function.
-
-    Returns:
-        Mapping from quantum operation name to its definition
-
-    """
-    op_table = copy.deepcopy(_default_operations)
-    for op_name, definition in ops.items():
-        implementations_def = definition.pop("implementations", {})
-        implementations = {}
-        for impl_name, class_name in implementations_def.items():
-            if (impl := _exposed_implementations.get(class_name)) is None:
-                raise ValueError(f"Implementation {class_name} has not been exposed.")
-            implementations[impl_name] = impl
-        if "defaults_for_locus" in definition:
-            for locus, impl_name in definition["defaults_for_locus"].items():
-                if impl_name not in implementations:
-                    raise ValueError(
-                        f"defaults_for_locus[{locus}] implementation '{impl_name}' does not "
-                        "appear in the implementations field of {op_name}."
-                    )
-        # assert that the attributes (other than default implementation info) are not changed:
-        if op_name in op_table:
-            for attribute_name, value in definition.items():
-                if not hasattr(op_table[op_name], attribute_name):
-                    raise ValueError(f"QuantumOp {op_name} has no attribute '{attribute_name}'.")
-                default_value = getattr(op_table[op_name], attribute_name)
-                if isinstance(value, Iterable):
-                    # cast all iterables to tuples for comparisons
-                    default_value = tuple(default_value)
-                    value = tuple(value)  # noqa: PLW2901
-                if attribute_name not in ("implementations", "defaults_for_locus") and default_value != value:
-                    raise ValueError(
-                        "The QuantumOp definition provided by the user has a different value in the"
-                        f" attribute {attribute_name} compared to the QuantumOp definition in iqm-pulse "
-                        f" ({value} != {default_value}). "
-                        "Changing the attributes of default QuantumOps is not allowed, other than"
-                        "implementations and defaults_for_locus."
-                    )
-            op_table[op_name] = replace(op_table[op_name], implementations=implementations, **definition)
-        else:
-            op_table[op_name] = QuantumOp(name=op_name, implementations=implementations, **definition)
-    return op_table
-
-
 def validate_quantum_circuit(
     operations: Iterable[CircuitOperation],
     op_table: QuantumOpTable,
@@ -228,6 +183,40 @@ def validate_quantum_circuit(
         raise ValueError("Circuit contains no measurements.")
 
 
+def build_quantum_ops(ops: dict[str, Any]) -> QuantumOpTable:
+    """Builds the table of known quantum operations.
+
+    Hardcoded default native ops table is extended by the ones in `ops`.
+    In case of name collisions, the content of `ops` takes priority over the defaults.
+
+    Args:
+        ops: Contents of the ``gate_definitions`` section defining
+        the quantum operations in the configuration YAML file.
+        Modified by the function.
+
+    Returns:
+        Mapping from quantum operation name to its definition
+
+    """
+    op_table = copy.deepcopy(_quantum_ops_library)
+    for op_name, definition in ops.items():
+        implementations_def = definition.pop("implementations", {})
+        implementations = _process_implementations(
+            op_name, implementations_def, _exposed_implementations, _default_implementations
+        )
+
+        if "defaults_for_locus" in definition:
+            _validate_locus_defaults(op_name, definition, implementations)
+
+        if op_name in op_table:
+            _validate_op_attributes(op_name, definition, op_table)
+            op_table[op_name] = replace(op_table[op_name], implementations=implementations, **definition)
+        else:
+            op_table[op_name] = QuantumOp(name=op_name, implementations=implementations, **definition)
+
+    return op_table
+
+
 class ScheduleBuilder:
     """Builds instruction schedules out of quantum circuits or individual quantum operations.
 
@@ -336,7 +325,7 @@ class ScheduleBuilder:
                         op in self._cache
                         and impl in self._cache[op]
                         and locus in self._cache[op][impl]
-                        and prev_calibration != new_calibration
+                        and _dicts_differ(prev_calibration, new_calibration)
                     ):
                         del self._cache[op][impl][locus]
                         if self.op_table[op].factorizable:
@@ -890,7 +879,7 @@ class ScheduleBuilder:
                 blocks only the defined locus components and any other components which have occupied channels.
 
         Returns:
-            playlist that implements ``boxes``, padded schedule duration in seconds
+            playlist that implements ``boxes``
 
         """
         schedules = [self.resolve_timebox(box, neighborhood=neighborhood).cleanup() for box in boxes]
@@ -975,15 +964,23 @@ class ScheduleBuilder:
 
         Args:
             schedule: schedule to finish
+
         Returns:
             finished copy of ``schedule``
 
         """
+        has_long_distance_rzs = False
 
-        def convert(inst: Instruction) -> Instruction:
-            """Convert Blocks and Nothings to Waits."""
+        def segment_pass(inst: Instruction) -> Instruction:
+            """Convert Blocks and Nothings to Waits. Also checks for FluxPulses that have long-distance VirtualZ
+            corrections in order to determine if these need handling (for performance reasons, we want to do this
+            step only on-demand).
+            """
+            nonlocal has_long_distance_rzs
             if isinstance(inst, NONSOLID) and not isinstance(inst, Wait):
                 return Wait(inst.duration)
+            if isinstance(inst, FluxPulse) and inst.rzs:
+                has_long_distance_rzs = True
             return inst
 
         def not_zero_block(inst: Instruction) -> bool:
@@ -1007,10 +1004,46 @@ class ScheduleBuilder:
         for channel in schedule.channels():
             seg = schedule[channel]
             schedule[channel] = Segment(
-                functools.reduce(merge_waits, map(convert, filter(not_zero_block, seg)), []),  # type: ignore[arg-type]
+                functools.reduce(merge_waits, map(segment_pass, filter(not_zero_block, seg)), []),
                 duration=seg.duration,
             )
-        return schedule.cleanup()
+        if not has_long_distance_rzs:
+            return schedule.cleanup()
+        return self._fuse_long_distance_virtual_rzs(schedule)
+
+    def _fuse_long_distance_virtual_rzs(self, schedule: Schedule) -> Schedule:
+        """Fuse long-distance (i.e. out-of-gate-locus) VirtualRZ corrections with the next drive pulse
+        happening after the FluxPulse they are correcting.
+
+        """
+
+        def fuse(inst: Instruction, idx: int, rz_angle: float, segment: Segment) -> None:
+            """Fuse a VirtualRZ angle to an IQPulse."""
+            if isinstance(inst, (VirtualRZ, IQPulse)):
+                segment._instructions[idx] = replace(inst, phase_increment=inst.phase_increment + rz_angle)
+            else:
+                raise ValueError(f"Unknown drive channel instruction {inst}")
+
+        schedule_copy = schedule.copy()
+        for channel in [ch for ch in schedule.channels() if "flux" in ch]:
+            sample_counter = 0
+            for inst in schedule[channel]:
+                if isinstance(inst, FluxPulse):
+                    rzs = {ch: angle for ch, angle in inst.rzs if ch in schedule}
+                    for drive_channel, rz_angle in rzs.items():
+                        drive_sample_counter = 0
+                        for drive_inst_idx, drive_inst in enumerate(schedule[drive_channel]):
+                            if not isinstance(drive_inst, Wait) and drive_sample_counter >= sample_counter:
+                                fuse(
+                                    schedule_copy[drive_channel][drive_inst_idx],
+                                    drive_inst_idx,
+                                    -rz_angle,  # invert the angle as per the agreed convention
+                                    schedule_copy[drive_channel],
+                                )
+                                break
+                            drive_sample_counter += drive_inst.duration
+                sample_counter += inst.duration
+        return schedule_copy.cleanup()
 
     def resolve_timebox(
         self, box: TimeBox, *, neighborhood: int, compute_neighborhood_hard_boundary: bool = False
